;; Generated package description from crisp.el  -*- no-byte-compile: t -*-
(define-package "crisp" "1.3.6.0.20201201.153359" "CRiSP/Brief Emacs emulator" 'nil :url "https://elpa.gnu.org/packages/crisp.html" :authors '(("Gary D. Foster" . "Gary.Foster@Corp.Sun.COM")) :maintainer '("Luke Lee" . "luke.yx.lee@gmail.com") :keywords '("emulations" "brief" "crisp"))
